"""Minidump processing"""
from __future__ import annotations

import shutil
from typing import IO

from symbolic._lowlevel import lib, ffi
from symbolic.utils import (
    RustObject,
    rustcall,
    encode_path,
    make_buffered_slice_reader,
)

__all__ = [
    "CfiCache",
    "CFICACHE_LATEST_VERSION",
]

# The most recent version for the CFI cache file format
CFICACHE_LATEST_VERSION = rustcall(lib.symbolic_cficache_latest_version)


class CfiCache(RustObject):
    """A cache for call frame information (CFI) to improve native stackwalking"""

    __dealloc_func__ = lib.symbolic_cficache_free

    @classmethod
    def open(cls, path: str) -> CfiCache:
        """Loads a cficache from a file via mmap."""
        return cls._from_objptr(rustcall(lib.symbolic_cficache_open, encode_path(path)))

    @classmethod
    def from_object(cls, obj: RustObject) -> CfiCache:
        """Creates a cficache from the given object."""
        return cls._from_objptr(
            rustcall(lib.symbolic_cficache_from_object, obj._get_objptr())
        )

    @property
    def version(self) -> str:
        """Version of the file format."""
        return self._methodcall(lib.symbolic_cficache_get_version)

    @property
    def is_latest_version(self) -> bool:
        """Returns true if this is the latest file format."""
        return self.version >= CFICACHE_LATEST_VERSION

    def open_stream(self) -> IO[bytes]:
        """Returns a stream to read files from the internal buffer."""
        buf = self._methodcall(lib.symbolic_cficache_get_bytes)
        size = self._methodcall(lib.symbolic_cficache_get_size)
        return make_buffered_slice_reader(ffi.buffer(buf, size), self)

    def write_to(self, f: IO[bytes]) -> None:
        """Writes the CFI cache into a file object."""
        # https://github.com/python/mypy/issues/15031
        shutil.copyfileobj(self.open_stream(), f)  # type: ignore[misc]



from __future__ import annotations

from symbolic._lowlevel import lib, ffi
from symbolic.utils import RustObject, rustcall, decode_str


__all__ = ["SourceMapCache", "SourceMapCacheToken"]


class SourceMapCacheToken:
    """Represents a token matched or looked up from the cache."""

    line: int
    col: int
    src: str | None
    name: str | None
    function_name: str | None
    context_line: str | None
    pre_context: list[str]
    post_context: list[str]

    @classmethod
    def _from_objptr(cls, tm: ffi.CData) -> SourceMapCacheToken:
        rv = object.__new__(cls)
        rv.line = tm.line
        rv.col = tm.col
        rv.src = decode_str(tm.src, free=False) or None
        rv.name = decode_str(tm.name, free=False) or None
        rv.function_name = decode_str(tm.function_name, free=False) or None

        rv.context_line = decode_str(tm.context_line, free=False) or None

        rv.pre_context = []
        for idx in range(tm.pre_context.len):
            s = decode_str(tm.pre_context.strs[idx], free=False)
            rv.pre_context.append(s)

        rv.post_context = []
        for idx in range(tm.post_context.len):
            s = decode_str(tm.post_context.strs[idx], free=False)
            rv.post_context.append(s)

        return rv

    def __repr__(self) -> str:
        return "<SourceMapCacheToken %s:%d>" % (
            self.src,
            self.line,
        )


class SourceMapCache(RustObject):
    """Gives access to a sm cache."""

    __dealloc_func__ = lib.symbolic_sourcemapcache_free

    @classmethod
    def from_bytes(
        cls, source_content: bytes, sourcemap_content: bytes
    ) -> SourceMapCache:
        """Constructs a sourcemapcache from bytes."""
        return cls._from_objptr(
            rustcall(
                lib.symbolic_sourcemapcache_from_bytes,
                source_content,
                len(source_content),
                sourcemap_content,
                len(sourcemap_content),
            )
        )

    def lookup(
        self, line: int, col: int, context_lines: int
    ) -> SourceMapCacheToken | None:
        """Looks up a token from the sourcemap."""
        rv = self._methodcall(
            lib.symbolic_sourcemapcache_lookup_token, line, col, context_lines
        )

        if rv != ffi.NULL:
            try:
                return SourceMapCacheToken._from_objptr(rv)
            finally:
                rustcall(lib.symbolic_sourcemapcache_token_match_free, rv)

        return None



__all__ = []


def _import_all():
    import pkgutil

    glob = globals()
    for _, modname, _ in pkgutil.iter_modules(__path__):
        if modname[:1] == "_":
            continue
        mod = __import__("symbolic.%s" % modname, glob, glob, ["__name__"])
        if not hasattr(mod, "__all__"):
            continue
        __all__.extend(mod.__all__)
        for name in mod.__all__:
            obj = getattr(mod, name)
            if hasattr(obj, "__module__"):
                obj.__module__ = "symbolic"
            glob[name] = obj


_import_all()
del _import_all



from __future__ import annotations

import io
import uuid
import ntpath
import weakref
import posixpath
from typing import Protocol
from symbolic._lowlevel import ffi, lib
from symbolic.exceptions import exceptions_by_code, SymbolicError


__all__ = ["common_path_join", "strip_common_path_prefix"]


attached_refs: weakref.WeakKeyDictionary[object, bytes]
attached_refs = weakref.WeakKeyDictionary()


def _is_win_path(x: str) -> bool:
    return "\\" in x or (ntpath.isabs(x) and not posixpath.isabs(x))


def common_path_join(a: str, b: str) -> str:
    """Joins two paths together while guessing the platform (win vs unix)."""
    if _is_win_path(a):
        return ntpath.normpath(ntpath.join(a, b))
    return posixpath.join(a, b)


class _PathMod(Protocol):
    sep: str

    def normpath(self, s: str) -> str:
        ...


def strip_common_path_prefix(base: str, prefix: str) -> str:
    """Strips `prefix` from `a`."""
    if _is_win_path(base):
        path: _PathMod = ntpath
    else:
        path = posixpath
    pieces_a = path.normpath(base).split(path.sep)
    pieces_b = path.normpath(prefix).split(path.sep)
    if pieces_a[: len(pieces_b)] == pieces_b:
        return path.sep.join(pieces_a[len(pieces_b) :])
    return path.normpath(base)


class RustObject:
    __dealloc_func__ = None
    _objptr = None
    _shared = False

    def __init__(self) -> None:
        raise TypeError("Cannot instanciate %r objects" % self.__class__.__name__)

    @classmethod
    def _from_objptr(cls, ptr, shared=False):
        rv = object.__new__(cls)
        rv._objptr = ptr
        rv._shared = shared
        return rv

    def _methodcall(self, func, *args):
        return rustcall(func, self._get_objptr(), *args)

    def _get_objptr(self):
        if not self._objptr:
            raise RuntimeError("Object is closed")
        return self._objptr

    def _move(self, target):
        self._shared = True
        ptr = self._get_objptr()
        self._objptr = None
        return ptr

    def __del__(self) -> None:
        if self._objptr is None or self._shared:
            return
        f = self.__class__.__dealloc_func__
        if f is not None:
            rustcall(f, self._objptr)
            self._objptr = None


def rustcall(func, *args):
    """Calls rust method and does some error handling."""
    lib.symbolic_err_clear()
    rv = func(*args)
    err = lib.symbolic_err_get_last_code()
    if not err:
        return rv
    msg = lib.symbolic_err_get_last_message()
    cls = exceptions_by_code.get(err, SymbolicError)
    exc = cls(decode_str(msg, free=True))
    backtrace = decode_str(lib.symbolic_err_get_backtrace(), free=True)
    if backtrace:
        exc.rust_info = backtrace
    raise exc


def decode_str(s: ffi.CData, free: bool = False) -> str:
    """Decodes a SymbolicStr"""
    try:
        if s.len == 0:
            return ""
        return ffi.unpack(s.data, s.len).decode("utf-8", "replace")
    finally:
        if free and s.owned:
            lib.symbolic_str_free(ffi.addressof(s))


def encode_str(s: str | bytes) -> ffi.CData:
    """Encodes a SymbolicStr"""
    rv = ffi.new("SymbolicStr *")
    if isinstance(s, str):
        s = s.encode("utf-8")
    rv.data = ffi.from_buffer(s)
    rv.len = len(s)
    # we have to hold a weak reference here to ensure our string does not
    # get collected before the string is used.
    attached_refs[rv] = s
    return rv


def encode_path(s: str | bytes) -> bytes:
    """Encodes a path value."""
    if isinstance(s, str):
        s = s.encode("utf-8")
    if 0 in s:
        raise TypeError("Null bytes are not allowed in paths")
    return s


def decode_uuid(value: ffi.CData) -> uuid.UUID:
    """Decodes the given uuid value."""
    return uuid.UUID(bytes=bytes(bytearray(ffi.unpack(value.data, 16))))


def encode_uuid(value: uuid.UUID | str) -> ffi.CData:
    """Encodes the given uuid value for FFI."""
    encoded = ffi.new("SymbolicUuid *")
    encoded.data[0:16] = bytearray(make_uuid(value).bytes)
    return encoded


def make_uuid(value: uuid.UUID | str) -> uuid.UUID:
    """Converts a value into a python uuid object."""
    if isinstance(value, uuid.UUID):
        return value
    return uuid.UUID(value)


class SliceReader(io.RawIOBase):
    """A buffered reader that keeps the cache in memory"""

    def __init__(self, buf, cache):
        self._buffer = buf
        # Hold the cache so we do not lose the reference and crash on
        # the buffer disappearing
        self.cache = cache
        self.pos = 0

    @property
    def size(self):
        return len(self._buffer)

    def readable(self):
        return True

    def readinto(self, buf):
        n = len(buf)
        if n is None:
            end = len(self._buffer)
        else:
            end = min(self.pos + n, len(self._buffer))
        rv = self._buffer[self.pos : end]
        buf[: len(rv)] = rv
        self.pos = end
        return len(rv)


class PassThroughBufferedReader(io.BufferedReader):
    __slots__ = ()

    def __getattr__(self, attr):
        return getattr(self.raw, attr)


def make_buffered_slice_reader(buffer, cache):
    return PassThroughBufferedReader(SliceReader(buffer, cache))



from __future__ import annotations
from typing import TYPE_CHECKING

from symbolic._lowlevel import lib


__all__ = ["SymbolicError"]
exceptions_by_code: dict[int, type[SymbolicError]] = {}


class SymbolicError(Exception):
    if TYPE_CHECKING:
        code: int
    else:
        code = None

    def __init__(self, msg):
        Exception.__init__(self)
        self.message = msg
        self.rust_info = None

    def __str__(self) -> str:
        rv = self.message
        if self.rust_info is not None:
            return f"{rv}\n\n{self.rust_info}"
        return rv


def _make_error(
    error_name: str, base: type[SymbolicError] = SymbolicError, code: int | None = None
) -> type[SymbolicError]:
    class Exc(base):  # type: ignore[misc,valid-type]
        pass

    Exc.__name__ = Exc.__qualname__ = error_name
    if code is not None:
        Exc.code = code
    globals()[Exc.__name__] = Exc
    __all__.append(Exc.__name__)
    return Exc


def _get_error_base(error_name: str) -> type[SymbolicError]:
    pieces = error_name.split("Error", 1)
    if len(pieces) == 2 and pieces[0] and pieces[1]:
        base_error_name = pieces[0] + "Error"
        base_class = globals().get(base_error_name)
        if base_class is None:
            base_class = _make_error(base_error_name)
        return base_class
    return SymbolicError


def _make_exceptions() -> None:
    for attr in dir(lib):
        if not attr.startswith("SYMBOLIC_ERROR_CODE_"):
            continue

        error_name = attr[20:].title().replace("_", "")
        base = _get_error_base(error_name)
        exc = _make_error(error_name, base=base, code=getattr(lib, attr))
        exceptions_by_code[exc.code] = exc


_make_exceptions()

if TYPE_CHECKING:
    # treat unknown attribute names as exception types
    def __getattr__(name: str) -> type[SymbolicError]:
        ...



from __future__ import annotations
from typing import Optional, Tuple

import uuid as uuid_mod

from symbolic._lowlevel import lib, ffi
from symbolic.utils import (
    RustObject,
    rustcall,
    decode_str,
    encode_str,
    decode_uuid,
    encode_path,
)


__all__ = ["ProguardMapper", "JavaStackFrame"]


class JavaStackFrame:
    def __init__(
        self,
        class_name: str,
        method: str,
        line: int,
        file: str | None = None,
        parameters: str | None = None,
    ) -> None:
        self.class_name = class_name
        self.method = method
        self.file = file or None
        self.line = line
        self.parameters = parameters


class ProguardMapper(RustObject):
    """Gives access to proguard mapping files."""

    __dealloc_func__ = lib.symbolic_proguardmapper_free

    @classmethod
    def open(cls, path: str, initialize_param_mapping: bool = False) -> ProguardMapper:
        """Constructs a mapping file from a path."""
        return cls._from_objptr(
            rustcall(
                lib.symbolic_proguardmapper_open,
                encode_path(path),
                initialize_param_mapping,
            )
        )

    @property
    def uuid(self) -> uuid_mod.UUID:
        """Returns the UUID of the file."""
        return decode_uuid(self._methodcall(lib.symbolic_proguardmapper_get_uuid))

    @property
    def has_line_info(self) -> bool:
        """True if the file contains line information."""
        return bool(self._methodcall(lib.symbolic_proguardmapper_has_line_info))

    def remap_class(self, klass: str) -> str | None:
        """Remaps the given class name."""
        klass = self._methodcall(
            lib.symbolic_proguardmapper_remap_class, encode_str(klass)
        )
        return decode_str(klass, free=True) or None

    def remap_method(self, klass: str, method: str) -> Tuple[str, str] | None:
        """Remaps the given class name."""
        result = self._methodcall(
            lib.symbolic_proguardmapper_remap_method,
            encode_str(klass),
            encode_str(method),
        )

        try:
            output = (
                decode_str(result.frames[0].class_name, free=False),
                decode_str(result.frames[0].method, free=False),
            )
        finally:
            rustcall(lib.symbolic_proguardmapper_result_free, ffi.addressof(result))

        return output if len(output[0]) > 0 and len(output[1]) > 0 else None

    def remap_frame(
        self,
        klass: str,
        method: str,
        line: int,
        parameters: Optional[str] = None,
    ) -> list[JavaStackFrame]:
        """Remaps the stackframe, given its class, method and line."""
        result = self._methodcall(
            lib.symbolic_proguardmapper_remap_frame,
            encode_str(klass),
            encode_str(method),
            line,
            encode_str("" if parameters is None else parameters),
            parameters is not None,
        )

        frames = []
        try:
            for idx in range(result.len):
                frame = result.frames[idx]
                frames.append(
                    JavaStackFrame(
                        decode_str(frame.class_name, free=False),
                        decode_str(frame.method, free=False),
                        frame.line,
                        decode_str(frame.file, free=False),
                    )
                )
        finally:
            rustcall(lib.symbolic_proguardmapper_result_free, ffi.addressof(result))

        return frames



from __future__ import annotations

from typing import Generator, overload

from symbolic._lowlevel import lib, ffi
from symbolic.utils import RustObject, rustcall, decode_str, encode_str, attached_refs


__all__ = ["SourceView", "SourceMapView", "SourceMapTokenMatch"]


class SourceMapTokenMatch:
    """Represents a token matched or looked up from the index."""

    def __init__(
        self,
        src_line: int,
        src_col: int,
        dst_line: int,
        dst_col: int,
        src_id: int | None = None,
        name: str | None = None,
        src: str | None = None,
        function_name: str | None = None,
    ):
        self.src_line = src_line
        self.src_col = src_col
        self.dst_line = dst_line
        self.dst_col = dst_col
        self.src_id = src_id
        self.name = name
        self.src = src
        self.function_name = function_name

    @classmethod
    def _from_objptr(cls, tm):
        return cls(
            src_line=tm.src_line,
            src_col=tm.src_col,
            dst_line=tm.dst_line,
            dst_col=tm.dst_col,
            src_id=tm.src_id,
            name=decode_str(tm.name, free=False) or None,
            src=decode_str(tm.src, free=False) or None,
            function_name=decode_str(tm.function_name, free=False) or None,
        )

    def __eq__(self, other: object) -> bool:
        if self.__class__ is not other.__class__:
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return "<SourceMapTokenMatch %s:%d>" % (
            self.src,
            self.src_line,
        )


class SourceView(RustObject):
    """Gives reasonably efficient access to javascript sourcecode."""

    __dealloc_func__ = lib.symbolic_sourceview_free

    @classmethod
    def from_bytes(cls, data: bytes) -> SourceView:
        """Constructs a source view from bytes."""
        data = bytes(data)
        rv = cls._from_objptr(
            rustcall(lib.symbolic_sourceview_from_bytes, data, len(data))
        )
        # we need to keep this reference alive or we crash. hard.
        attached_refs[rv] = data
        return rv

    def get_source(self) -> str:
        source = self._methodcall(lib.symbolic_sourceview_as_str)
        return decode_str(source, free=True)

    def __len__(self) -> int:
        return self._methodcall(lib.symbolic_sourceview_get_line_count)

    @overload
    def __getitem__(self, idx: int) -> str:
        ...

    @overload
    def __getitem__(self, idx: slice) -> list[str]:
        ...

    def __getitem__(self, idx: int | slice) -> str | list[str]:
        if not isinstance(idx, slice):
            if idx >= len(self):
                raise IndexError("No such line")
            line = self._methodcall(lib.symbolic_sourceview_get_line, idx)
            return decode_str(line, free=True)

        rv = []
        for idx in range(*idx.indices(len(self))):
            try:
                rv.append(self[idx])
            except IndexError:
                pass
        return rv

    def __iter__(self) -> Generator[str, None, None]:
        for x in range(len(self)):
            yield self[x]


class SourceMapView(RustObject):
    """Gives access to a source map."""

    __dealloc_func__ = lib.symbolic_sourcemapview_free

    @classmethod
    def from_json_bytes(cls, data: bytes) -> SourceMapView:
        """Constructs a sourcemap from bytes of JSON data."""
        data = bytes(data)
        return cls._from_objptr(
            rustcall(lib.symbolic_sourcemapview_from_json_slice, data, len(data))
        )

    def lookup(
        self,
        line: int,
        col: int,
        minified_function_name: str | None = None,
        minified_source: SourceView | None = None,
    ) -> SourceMapTokenMatch | None:
        """Looks up a token from the sourcemap and optionally also
        resolves a function name from a stacktrace to the original one.
        """
        if minified_function_name is None or minified_source is None:
            rv = self._methodcall(lib.symbolic_sourcemapview_lookup_token, line, col)
        else:
            if not isinstance(minified_source, SourceView):
                raise TypeError("source view required")
            rv = self._methodcall(
                lib.symbolic_sourcemapview_lookup_token_with_function_name,
                line,
                col,
                encode_str(minified_function_name),
                minified_source._objptr,
            )
        if rv != ffi.NULL:
            try:
                return SourceMapTokenMatch._from_objptr(rv)
            finally:
                rustcall(lib.symbolic_token_match_free, rv)

        return None

    def get_sourceview(self, idx: int) -> SourceView | None:
        """Given a source index returns the source view that created it."""
        rv = self._methodcall(lib.symbolic_sourcemapview_get_sourceview, idx)
        if rv != ffi.NULL:
            return SourceView._from_objptr(rv, shared=True)

        return None

    @property
    def source_count(self) -> int:
        """Returns the number of sources."""
        return self._methodcall(lib.symbolic_sourcemapview_get_source_count)

    def get_source_name(self, idx: int) -> str | None:
        """Returns the name of the source at the given index."""
        name = self._methodcall(lib.symbolic_sourcemapview_get_source_name, idx)
        return decode_str(name, free=True) or None

    def get_source_contents(self, idx: int) -> str | None:
        """Returns the contents of the source at the given index."""
        name = self._methodcall(lib.symbolic_sourcemapview_get_source_contents, idx)
        return decode_str(name, free=True) or None

    def iter_sources(self) -> Generator[tuple[int, str | None], None, None]:
        """Iterates over the sources in the file."""
        for src_id in range(self.source_count):
            yield src_id, self.get_source_name(src_id)

    def __len__(self) -> int:
        return self._methodcall(lib.symbolic_sourcemapview_get_tokens)

    def __getitem__(self, idx: int) -> SourceMapTokenMatch:
        rv = self._methodcall(lib.symbolic_sourcemapview_get_token, idx)
        if rv == ffi.NULL:
            raise IndexError("Token out of range")
        try:
            return SourceMapTokenMatch._from_objptr(rv)
        finally:
            rustcall(lib.symbolic_token_match_free, rv)

    def __iter__(self) -> Generator[SourceMapTokenMatch, None, None]:
        for x in range(len(self)):
            yield self[x]



from __future__ import annotations

import os
from typing import overload

from symbolic._lowlevel import lib, ffi
from symbolic.utils import rustcall, encode_str, decode_str
from symbolic import exceptions


__all__ = ["arch_is_known", "arch_get_ip_reg_name", "normalize_arch", "parse_addr"]


ignore_arch_exc = (exceptions.UnknownArchError,)


# Make sure we init the lib and turn on rust backtraces
os.environ["RUST_BACKTRACE"] = "1"
ffi.init_once(lib.symbolic_init, "init")


def arch_is_known(arch: str) -> bool:
    """Checks if an architecture is known."""
    if not isinstance(arch, str):
        return False
    return rustcall(lib.symbolic_arch_is_known, encode_str(arch))


@overload
def normalize_arch(arch: None) -> None:
    ...


@overload
def normalize_arch(arch: str) -> str:
    ...


def normalize_arch(arch: str | None) -> str | None:
    """Normalizes an architecture name."""
    if arch is None:
        return None
    if not isinstance(arch, str):
        raise ValueError("Invalid architecture: expected string")

    normalized = rustcall(lib.symbolic_normalize_arch, encode_str(arch))
    return decode_str(normalized, free=True)


def arch_get_ip_reg_name(arch: str) -> str | None:
    """Returns the ip register if known for this arch."""
    try:
        rv = rustcall(lib.symbolic_arch_ip_reg_name, encode_str(arch))
        return str(decode_str(rv, free=True))
    except ignore_arch_exc:
        return None


def parse_addr(x: int | str | None) -> int:
    """Parses an address."""
    if x is None:
        return 0
    if isinstance(x, int):
        return x
    if isinstance(x, str):
        if x[:2] == "0x":
            return int(x[2:], 16)
        return int(x)
    raise ValueError(f"Unsupported address format {x!r}")



from __future__ import annotations

import bisect
from weakref import WeakValueDictionary
from typing import overload, Generator

from symbolic._lowlevel import lib, ffi
from symbolic.utils import RustObject, rustcall, decode_str, encode_str
from symbolic.common import parse_addr, arch_is_known
from symbolic.symcache import SymCache
from symbolic.cfi import CfiCache


__all__ = [
    "Archive",
    "Object",
    "ObjectLookup",
    "BcSymbolMap",
    "UuidMapping",
    "id_from_breakpad",
    "normalize_code_id",
    "normalize_debug_id",
]


class Archive(RustObject):
    __dealloc_func__ = lib.symbolic_archive_free

    _objcache: WeakValueDictionary[int, Object]

    @classmethod
    def open(self, path: str | bytes) -> Archive:
        """Opens an archive from a given path."""
        if isinstance(path, str):
            path = path.encode("utf-8")
        return Archive._from_objptr(rustcall(lib.symbolic_archive_open, path))

    @classmethod
    def from_bytes(self, data: bytes) -> Archive:
        """Loads an archive from a binary buffer."""
        return Archive._from_objptr(
            rustcall(lib.symbolic_archive_from_bytes, data, len(data))
        )

    @property
    def object_count(self) -> int:
        """The number of objects in this archive."""
        return self._methodcall(lib.symbolic_archive_object_count)

    def iter_objects(self) -> Generator[Object, None, None]:
        """Iterates over all objects."""
        for idx in range(self.object_count):
            try:
                yield self._get_object(idx)
            except LookupError:
                pass

    def get_object(
        self, debug_id: str | None = None, arch: str | None = None
    ) -> Object:
        """Get an object by either arch or id."""
        if debug_id is not None:
            debug_id = debug_id.lower()
        for obj in self.iter_objects():
            if obj.debug_id == debug_id or obj.arch == arch:
                return obj
        raise LookupError("Object not found")

    def _get_object(self, idx: int) -> Object:
        """Returns the object at a certain index."""
        cache = getattr(self, "_objcache", None)
        if cache is None:
            cache = self._objcache = WeakValueDictionary()
        rv = cache.get(idx)
        if rv is not None:
            return rv
        ptr = self._methodcall(lib.symbolic_archive_get_object, idx)
        if ptr == ffi.NULL:
            raise LookupError("No object #%d" % idx)
        rv = cache[idx] = Object._from_objptr(ptr)
        return rv


class Object(RustObject):
    __dealloc_func__ = lib.symbolic_object_free

    @property
    def arch(self) -> str:
        """The architecture of the object."""
        # make it an ascii bytestring on 2.x
        arch = self._methodcall(lib.symbolic_object_get_arch)
        return str(decode_str(arch, free=True))

    @property
    def code_id(self) -> str | None:
        """The code identifier of the object. Returns None if there is no code id."""
        code_id = self._methodcall(lib.symbolic_object_get_code_id)
        code_id = decode_str(code_id, free=True)
        if code_id:
            return code_id
        return None

    @property
    def debug_id(self) -> str:
        """The debug identifier of the object."""
        debug_id = self._methodcall(lib.symbolic_object_get_debug_id)
        return decode_str(debug_id, free=True)

    @property
    def kind(self) -> str:
        """The kind of the object (e.g. executable, debug file, library, ...)."""
        kind = self._methodcall(lib.symbolic_object_get_kind)
        return str(decode_str(kind, free=True))

    @property
    def file_format(self) -> str:
        """The file format of the object file (e.g. MachO, ELF, ...)."""
        format = self._methodcall(lib.symbolic_object_get_file_format)
        return str(decode_str(format, free=True))

    @property
    def features(self) -> frozenset[str]:
        """The list of features offered by this debug file."""
        struct = self._methodcall(lib.symbolic_object_get_features)
        features = set()
        if struct.symtab:
            features.add("symtab")
        if struct.debug:
            features.add("debug")
        if struct.unwind:
            features.add("unwind")
        if struct.sources:
            features.add("sources")
        return frozenset(features)

    def make_symcache(self) -> SymCache:
        """Creates a symcache from the object."""
        return SymCache._from_objptr(
            self._methodcall(lib.symbolic_symcache_from_object)
        )

    def make_cficache(self) -> CfiCache:
        """Creates a cficache from the object."""
        return CfiCache._from_objptr(
            self._methodcall(lib.symbolic_cficache_from_object)
        )

    def __repr__(self) -> str:
        return "<Object {} {!r}>".format(
            self.debug_id,
            self.arch,
        )


class ObjectRef:
    """Holds a reference to an object in a format."""

    def __init__(self, data):
        self.addr = parse_addr(data.get("image_addr"))
        # not a real address but why handle it differently
        self.size = parse_addr(data.get("image_size"))
        self.vmaddr = data.get("image_vmaddr")
        self.code_id = data.get("code_id")
        self.code_file = data.get("code_file") or data.get("name")
        self.debug_id = normalize_debug_id(
            data.get("debug_id") or data.get("id") or data.get("uuid") or None
        )
        self.debug_file = data.get("debug_file")

        if data.get("arch") is not None and arch_is_known(data["arch"]):
            self.arch = data["arch"]
        else:
            self.arch = None

        # Legacy alias for backwards compatibility
        self.name = self.code_file

    def __repr__(self) -> str:
        return "<ObjectRef {} {!r}>".format(
            self.debug_id,
            self.arch,
        )


class ObjectLookup:
    """Helper to look up objects based on the info a client provides."""

    def __init__(self, objects):
        self._addresses = []
        self._by_addr = {}
        self.objects = {}
        for ref_data in objects:
            obj = ObjectRef(ref_data)
            self._addresses.append(obj.addr)
            self._by_addr[obj.addr] = obj
            self.objects[obj.debug_id] = obj
        self._addresses.sort()

    def iter_objects(self):
        """Iterates over all objects."""
        return self.objects.values()

    def get_debug_ids(self):
        """Returns a list of ids."""
        return sorted(self.objects)

    def iter_debug_ids(self):
        """Iterates over all ids."""
        return iter(self.objects)

    def find_object(self, addr):
        """Given an instruction address this locates the image this address
        is contained in.
        """
        idx = bisect.bisect_right(self._addresses, parse_addr(addr))
        if idx > 0:
            rv = self._by_addr[self._addresses[idx - 1]]
            if not rv.size or parse_addr(addr) < rv.addr + rv.size:
                return rv

    def get_object(self, debug_id):
        """Finds an object by the given debug id."""
        return self.objects.get(debug_id)


class BcSymbolMap(RustObject):
    """Object representing an Apple ``.bcsymbolmap`` file."""

    __dealloc_func__ = lib.symbolic_bcsymbolmap_free

    @classmethod
    def open(cls, path: str | bytes) -> BcSymbolMap:
        """Parses a BCSymbolMap file."""
        if isinstance(path, str):
            path = path.encode("utf-8")
        return cls._from_objptr(rustcall(lib.symbolic_bcsymbolmap_open, path))


class UuidMapping(RustObject):
    """Object representing a mapping from one DebugID to another."""

    __dealloc_func__ = lib.symbolic_uuidmapping_free

    @classmethod
    def from_plist(cls, debug_id: str, path: str | bytes) -> UuidMapping:
        """Parses a PList."""
        if isinstance(path, str):
            path = path.encode("utf-8")
        return cls._from_objptr(
            rustcall(lib.symbolic_uuidmapping_from_plist, encode_str(debug_id), path)
        )


def id_from_breakpad(breakpad_id):
    """Converts a Breakpad CodeModuleId to DebugId"""
    if breakpad_id is None:
        return None

    s = encode_str(breakpad_id)
    id = rustcall(lib.symbolic_id_from_breakpad, s)
    return decode_str(id, free=True)


def normalize_code_id(code_id):
    """Normalizes a code identifier to default representation"""
    if code_id is None:
        return None

    s = encode_str(code_id)
    id = rustcall(lib.symbolic_normalize_code_id, s)
    return decode_str(id, free=True)


@overload
def normalize_debug_id(debug_id: None) -> None:
    ...


@overload
def normalize_debug_id(debug_id: str) -> str:
    ...


def normalize_debug_id(debug_id: str | None) -> str | None:
    """Normalizes a debug identifier to default representation"""
    if debug_id is None:
        return None

    s = encode_str(debug_id)
    id = rustcall(lib.symbolic_normalize_debug_id, s)
    return decode_str(id, free=True)



import os
import re
import sys
import atexit
import shutil
import zipfile
import tempfile
import subprocess
from setuptools import setup, find_packages
from setuptools.command.sdist import sdist


_version_re = re.compile(r'(?m)^version\s*=\s*"(.*?)"\s*$')


DEBUG_BUILD = os.environ.get("SYMBOLIC_DEBUG") == "1"

with open("README") as f:
    readme = f.read()


if os.path.isfile("../symbolic-cabi/Cargo.toml"):
    with open("../symbolic-cabi/Cargo.toml") as f:
        match = _version_re.search(f.read())
        assert match is not None
        version = match[1]
else:
    with open("version.txt") as f:
        version = f.readline().strip()


def vendor_rust_deps():
    subprocess.Popen(["scripts/git-archive-all", "py/rustsrc.zip"], cwd="..").wait()


def write_version():
    with open("version.txt", "w") as f:
        f.write("%s\n" % version)


class CustomSDist(sdist):
    def run(self):
        vendor_rust_deps()
        write_version()
        sdist.run(self)


def build_native(spec):
    cmd = ["cargo", "build", "-p", "symbolic-cabi"]
    if not DEBUG_BUILD:
        cmd.append("--release")
        target = "release"
    else:
        target = "debug"

    # Step 0: find rust sources
    if not os.path.isfile("../symbolic-cabi/Cargo.toml"):
        scratchpad = tempfile.mkdtemp()

        @atexit.register
        def delete_scratchpad():
            try:
                shutil.rmtree(scratchpad)
            except OSError:
                pass

        zf = zipfile.ZipFile("rustsrc.zip")
        zf.extractall(scratchpad)
        rust_path = scratchpad + "/rustsrc"
    else:
        rust_path = ".."
        scratchpad = None

    # Step 1: build the rust library
    print("running `{}` ({} target)".format(" ".join(cmd), target))
    build = spec.add_external_build(cmd=cmd, path=rust_path)

    def find_dylib():
        cargo_target = os.environ.get("CARGO_BUILD_TARGET")
        if cargo_target:
            in_path = f"target/{cargo_target}/{target}"
        else:
            in_path = "target/%s" % target
        return build.find_dylib("symbolic_cabi", in_path=in_path)

    rtld_flags = ["NOW"]
    if sys.platform == "darwin":
        rtld_flags.append("NODELETE")
    spec.add_cffi_module(
        module_path="symbolic._lowlevel",
        dylib=find_dylib,
        header_filename=lambda: build.find_header(
            "symbolic.h", in_path="symbolic-cabi/include"
        ),
        rtld_flags=rtld_flags,
    )


setup(
    name="symbolic",
    version=version,
    packages=find_packages(),
    author="Sentry",
    license="MIT",
    author_email="hello@sentry.io",
    description="A python library for dealing with symbol files and more.",
    long_description=readme,
    include_package_data=True,
    package_data={"symbolic": ["py.typed", "_lowlevel.pyi"]},
    long_description_content_type="text/markdown",
    zip_safe=False,
    platforms="any",
    install_requires=["milksnake>=0.1.2"],
    # Specify transitive dependencies manually that are required for the build because
    # they are not resolved properly since upgrading to python 3.11 in manylinux.
    # milksnake -> cffi -> pycparser
    # milksnake specifies cffi>=1.6.0 as dependency while cffi does not specify a
    # minimum version for pycparser
    setup_requires=["milksnake>=0.1.2", "cffi>=1.6.0", "pycparser"],
    python_requires=">=3.8",
    milksnake_tasks=[build_native],
    cmdclass={"sdist": CustomSDist},
)



import io
import shutil
from symbolic._lowlevel import lib, ffi
from symbolic.utils import (
    RustObject,
    rustcall,
    decode_str,
    encode_str,
    encode_path,
    attached_refs,
    SliceReader,
)
from symbolic.common import parse_addr
from symbolic import exceptions


__all__ = [
    "SourceLocation",
    "SymCache",
    "find_best_instruction",
    "SYMCACHE_LATEST_VERSION",
]


# the most recent version for the symcache file format.
SYMCACHE_LATEST_VERSION = rustcall(lib.symbolic_symcache_latest_version)


class SourceLocation:
    def __init__(self, sym_addr, instr_addr, line, lang, symbol, full_path=None):
        self.sym_addr = sym_addr
        self.instr_addr = instr_addr
        self.line = line
        self.lang = lang
        self.symbol = symbol
        self.full_path = full_path or None

    def __str__(self) -> str:
        return "{}:{} ({})".format(
            self.symbol,
            self.line,
            self.full_path,
        )

    def __repr__(self) -> str:
        return "SourceLocation(%s)" % (
            ", ".join("%s=%r" % x for x in sorted(self.__dict__.items()))
        )


class SymCache(RustObject):
    __dealloc_func__ = lib.symbolic_symcache_free

    @classmethod
    def open(cls, path):
        """Loads a symcache from a file via mmap."""
        return cls._from_objptr(rustcall(lib.symbolic_symcache_open, encode_path(path)))

    @classmethod
    def from_object(cls, obj):
        """Creates a symcache from the given object."""
        return cls._from_objptr(
            rustcall(lib.symbolic_symcache_from_object, obj._get_objptr())
        )

    @classmethod
    def from_bytes(cls, data):
        """Loads a symcache from a binary buffer."""
        symcache = cls._from_objptr(
            rustcall(lib.symbolic_symcache_from_bytes, data, len(data))
        )
        attached_refs[symcache] = data
        return symcache

    @property
    def arch(self):
        """The architecture of the symcache."""
        arch = self._methodcall(lib.symbolic_symcache_get_arch)
        # make it an ascii bytestring on 2.x
        return str(decode_str(arch, free=True))

    @property
    def debug_id(self):
        """The debug identifier of the object."""
        id = self._methodcall(lib.symbolic_symcache_get_debug_id)
        return decode_str(id, free=True)

    @property
    def version(self):
        """Version of the file format."""
        return self._methodcall(lib.symbolic_symcache_get_version)

    @property
    def is_latest_version(self):
        """Returns true if this is the latest file format."""
        return self.version >= SYMCACHE_LATEST_VERSION

    def open_stream(self):
        """Returns a stream to read files from the internal buffer."""
        buf = self._methodcall(lib.symbolic_symcache_get_bytes)
        size = self._methodcall(lib.symbolic_symcache_get_size)
        return io.BufferedReader(SliceReader(ffi.buffer(buf, size), self))

    def dump_into(self, f):
        """Dumps the symcache into a file object."""
        shutil.copyfileobj(self.open_stream(), f)

    def lookup(self, addr):
        """Look up a single address."""
        addr = parse_addr(addr)
        rv = self._methodcall(lib.symbolic_symcache_lookup, addr)
        try:
            matches = []
            for idx in range(rv.len):
                sym = rv.items[idx]
                lang = decode_str(sym.lang, free=False)
                symbol = decode_str(sym.symbol, free=False)
                full_path = decode_str(sym.full_path, free=True)
                matches.append(
                    SourceLocation(
                        sym_addr=sym.sym_addr,
                        instr_addr=sym.instr_addr,
                        line=sym.line,
                        lang=lang,
                        symbol=symbol,
                        full_path=full_path,
                    )
                )
        finally:
            rustcall(lib.symbolic_lookup_result_free, ffi.addressof(rv))
        return matches


def find_best_instruction(addr, arch, crashing_frame=False, signal=None, ip_reg=None):
    """Given an instruction and meta data attempts to find the best one
    by using a heuristic we inherited from symsynd.
    """
    # Ensure we keep this local alive until this function returns as we
    # would otherwise operate on garbage
    encoded_arch = encode_str(arch)

    addr = parse_addr(addr)
    ii = ffi.new("SymbolicInstructionInfo *")
    ii[0].addr = addr
    ii[0].arch = encoded_arch
    ii[0].crashing_frame = crashing_frame
    ii[0].signal = signal or 0
    ii[0].ip_reg = parse_addr(ip_reg)
    try:
        return int(rustcall(lib.symbolic_find_best_instruction, ii))
    except exceptions.UnknownArchError:
        return int(addr)
